/*
 * Copyright (C) 2010  Michael Morin
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the  v. 3.0  or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the Academic Free License ("AFL")
 * v. 3.0  for more details.
 *
 * You should have received a copy of the Academic Free License ("AFL") v. 3.0
 * along with this program.
 *
 */

/*
 * From the author:
 * This software uses the following third party libraries:
 *  - Choco Solver 2.1.3 (http://www.emn.fr/z-info/choco-solver)
 *  - Jung 2.0.1 (http://jung.sourceforge.net)
 * and is related to the following publication:
 *  M. Morin, A.P. Papillon, F. Laviolette, I. Abi-Zeid, and C.G. Quimper, 
 *  "Constraint Programming for Path Planning with Uncertainty: Solving the 
 *  Optimal Search Path problem," in Proceedings of the 18th Conference on 
 *  Principles and Practice of Constraint Programming, Qubec, Qc, Canada, 
 *  2012, pp. 988-1003.
 * 
 * Enhancements and comments are welcomed. Please let us know!  
 *
 * Best regards,
 *
 * Michael
 * http://www.michaelmorin.info
 */

/** 
 * @file DynamicTotalDetectionValIterator.java
 * @brief The DynamicTotalDetectionValIterator class.
 */

package ospCp2012;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.TreeMap;

import choco.kernel.solver.search.ValIterator;
import choco.kernel.solver.variables.integer.IntDomainVar;
import edu.uci.ics.jung.graph.Graph;

/**
 * @class DynamicTotalDetectionValIterator
 * @version 1.0
 * @author Michael Morin
 * @date Created on: 2012-04-20
 * @date Modified on: 2012-05-25
 *  
 * @brief  class DynamicTotalDetectionValIterator
 *
 * This class implements the total detection (TD) value selection heuristic
 * for the Optimal Search Path (OSP) problem.
 * 
 */
public class DynamicTotalDetectionValIterator implements ValIterator<IntDomainVar> {
	protected static final double FLOAT_TOLERANCE = Double.MIN_VALUE;
	protected static final int FIRST_READING_POSITION_IN_Y_NAME = 2;
	
	protected final double[][][] mW;
	protected final IntDomainVar[] mSearchPath;
	protected final IntDomainVar[][] mPoc;
	protected final ArrayList< ArrayList< ArrayList<Integer> > > mFasOrdering;
	protected int[] mCurrentFasOrdering;
	protected final int[][] mFasYPointer;
	protected final int mT;
	protected final int mN;
	protected final Graph<Integer,Integer> mEnv;
	protected final double[][] mMotionModel;
	protected final double mProbabilityOfDetection;
	
	public DynamicTotalDetectionValIterator(IntDomainVar[] searchPath, IntDomainVar[] poc, int T, int N, Graph<Integer, Integer> env, double[][] motionModel, double pod) {
		//searchPath is assumed to be sorted in ascending order of t
		super();
		
		mSearchPath = searchPath;
		
		mFasOrdering = new ArrayList< ArrayList< ArrayList<Integer> > >(T);

		for(int t = 0; t < T; ++t) {
			mFasOrdering.add(new ArrayList< ArrayList<Integer> >(N));
			for(int y = 0; y < N; ++y) {
				mFasOrdering.get(mFasOrdering.size() - 1).add(new ArrayList<Integer>());
			}
		}
		
		mFasYPointer = new int[T][N];
		for(int t = 0; t < T; ++t) {
			for(int y = 0; y < N; ++y) {
				mFasYPointer[t][y] = -2;
			}
		}
		
		mT = T;
		mN = N;
		mEnv = env;
		mPoc = new IntDomainVar[T][N];
		for(int t = 0; t < T; ++t) {
			for(int y = 0; y < N; ++y) {
				mPoc[t][y] = poc[t * N + y];
			}
		}
		
		mMotionModel = motionModel;
		mProbabilityOfDetection = pod;
		mW = computeW();
	}
	
	/**
	* testing whether more branches can be considered after branch i, on the alternative
	associated to variable x
	*
	* @param y the variable under scrutiny
	* @param i the index of the last branch explored
	* @return true if more branches can be expanded after branch i
	*/
	public boolean hasNextVal(IntDomainVar y, int i) {
		final int t = Integer.parseInt(y.getName().substring(FIRST_READING_POSITION_IN_Y_NAME,y.getName().indexOf(")")));
		return (mFasYPointer[t - 1][mSearchPath[t - 1].getVal()] < mFasOrdering.get(t - 1).get(mSearchPath[t - 1].getVal()).size());
	}
	/**
	* Accessing the index of the first branch for variable y
	*
	* @param y the variable under scrutiny
	* @return the index of the first branch (such as the first value to be assigned to the
	variable)
	*/
	public int getFirstVal(IntDomainVar y) {
		final int t = Integer.parseInt(y.getName().substring(FIRST_READING_POSITION_IN_Y_NAME,y.getName().indexOf(")")));	
		
		mFasYPointer[t - 1][mSearchPath[t - 1].getVal()] = 0;
		mFasOrdering.get(t - 1).set(mSearchPath[t - 1].getVal(), computeSolutionsOrderedSetUsingExpectedDetectionByTimeSteps(mW, mPoc, t - 1, mSearchPath[t - 1].getVal()));
		while(y.canBeInstantiatedTo(mFasOrdering.get(t - 1).get(mSearchPath[t - 1].getVal()).get(mFasYPointer[t - 1][mSearchPath[t - 1].getVal()])) == false) {
			mFasYPointer[t - 1][mSearchPath[t - 1].getVal()] += 1;
		}
		
		return mFasOrdering.get(t - 1).get(mSearchPath[t - 1].getVal()).get(mFasYPointer[t - 1][mSearchPath[t - 1].getVal()]);
	}
	/**
	* Generates the index of the next branch after branch i, on the alternative associated to
	* variable y
	*
	* @param y the variable under scrutiny
	* @param i the index of the last branch explored
	* @return the index of the next branch to be expanded after branch i
	*/
	public int getNextVal(IntDomainVar y, int i) {
		final int t = Integer.parseInt(y.getName().substring(FIRST_READING_POSITION_IN_Y_NAME,y.getName().indexOf(")")));

		mFasYPointer[t - 1][mSearchPath[t - 1].getVal()] += 1;
		while(y.canBeInstantiatedTo(mFasOrdering.get(t - 1).get(mSearchPath[t - 1].getVal()).get(mFasYPointer[t - 1][mSearchPath[t - 1].getVal()])) == false) {
			mFasYPointer[t - 1][mSearchPath[t - 1].getVal()] += 1;
		}
		return mFasOrdering.get(t - 1).get(mSearchPath[t - 1].getVal()).get(mFasYPointer[t - 1][mSearchPath[t - 1].getVal()]);
	}
	
	/**
	 * Compute the matrix W containing the probability of further detection.
	 * 
	 * @return The matrix W containing the probability of further object's detection.
	 */
	private double[][][] computeW() {
		
		double[][][] var_w = new double[mT][mN][mN];
		
		for(int y = 0; y < mN; ++y) {
			for(int o = 0; o < mN; ++o) {
				if(y == o) {
					var_w[mT - 1][y][o] = mProbabilityOfDetection;
				} else {
					var_w[mT - 1][y][o] = 0.0;
				}
			}
		}
		
		for(int t = mT - 2; t > -1; --t) {
			for(int o = 0; o < mN; ++o) {
				for(int y = 0; y < mN; ++y) {
					var_w[t][y][o] = 0;
					for(int o_p = 0; o_p < mN; ++o_p) {
						if(mEnv.findEdge(o, o_p) != null) {
							double max_w = Double.MIN_VALUE;
							for(int y_p = 0; y_p < mN; ++y_p) {
								if((var_w[t+1][y_p][o_p] > max_w) && (mEnv.findEdge(y,y_p) != null)){
									max_w = var_w[t+1][y_p][o_p];
								}
							}
							var_w[t][y][o] += mMotionModel[o][o_p] * max_w;
						}
					}
					if (o == y) {
						var_w[t][y][o] = mProbabilityOfDetection + (1.0 - mProbabilityOfDetection) * var_w[t][y][o];
					}
				}
			}
		}
		
		return var_w;
	}
	
	private ArrayList<Integer> computeSolutionsOrderedSetUsingExpectedDetectionByTimeSteps(double[][][] varW, IntDomainVar[][] poc, int t, int y) {
		assert(t < mT);
		
		TreeMap<Double,ArrayList<Integer>> ordering = new TreeMap<Double,ArrayList<Integer>>();
			// The ordering at time t for a region y. 
		int added_region_count = 0;

		while(added_region_count < mEnv.getNeighborCount(y)) {
			Iterator<Integer> iter_neighbors_y = mEnv.getNeighbors(y).iterator();
			while(iter_neighbors_y.hasNext()) {
				int y_p = iter_neighbors_y.next();
				Double mean_w = 0.0;
				for(int o = 0; o < mN; ++o) {
					mean_w += varW[t][y_p][o] * poc[t][o].getVal();
				}
				if(ordering.containsKey(mean_w)) {
					ordering.get(mean_w).add(y_p);
					added_region_count++;
				} else {
					ordering.put(mean_w, new ArrayList<Integer>(mEnv.getNeighbors(y).size()));
					ordering.get(mean_w).add(y_p);
					added_region_count++;
				}
			}
		}
		
		ArrayList<Integer> ordered_regions = new ArrayList<Integer>(mEnv.getNeighborCount(y));
		Iterator<Double> key_iter = ordering.descendingKeySet().iterator();
		int i = 0;
		while(key_iter.hasNext()) {
			Iterator<Integer> entries_iter = ordering.get(key_iter.next()).iterator();
			while(entries_iter.hasNext()) {
				ordered_regions.add(entries_iter.next());
				i++;
			}
		}
		
		return ordered_regions; 
	}

}

//End of file
