/*
 * Copyright (C) 2010  Michael Morin
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the  v. 3.0  or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the Academic Free License ("AFL")
 * v. 3.0  for more details.
 *
 * You should have received a copy of the Academic Free License ("AFL") v. 3.0
 * along with this program.
 *
 */

/*
 * From the author:
 * This software uses the following third party libraries:
 *  - Choco Solver 2.1.3 (http://www.emn.fr/z-info/choco-solver)
 *  - Jung 2.0.1 (http://jung.sourceforge.net)
 * and is related to the following publication:
 *  M. Morin, A.P. Papillon, F. Laviolette, I. Abi-Zeid, and C.G. Quimper, 
 *  "Constraint Programming for Path Planning with Uncertainty: Solving the 
 *  Optimal Search Path problem," in Proceedings of the 18th Conference on 
 *  Principles and Practice of Constraint Programming, Qubec, Qc, Canada, 
 *  2012, pp. 988-1003.
 * 
 * Enhancements and comments are welcomed. Please let us know! 
 *
 * Best regards,
 *
 * Michael
 * http://www.michaelmorin.info
 */

/** 
 * @file InstanceConfig.java
 * @brief The InstanceConfig class.
 */

package ospCp2012;

import edu.uci.ics.jung.graph.Graph;

/**
 * @class InstanceConfig
 * @version 1.0
 * @author Michael Morin
 * @date Created on: 2011-04-10
 * @date Modified on: 2012-04-02
 *  
 * @brief  class InstanceConfig
 *
 * This class is a problem instance generator based on the instance
 * identifier. 
 * <p>
 * Each instance id has six digit. The first digit determines the 
 * accessibility graph type. The second digit determines the
 * probability of detection for all vertices of the accessibility graph.
 * The third digit determine the Markovian motion model of the object.
 * The last three digits specify the path length, i.e., the total number
 * of time steps that are available to search the accessibility graph.
 *  
 */
public class InstanceConfig {
	
	protected static final int OSP_ENVIRONMENT_CAMPUS = 1;
	protected static final int OSP_ENVIRONMENT_PLUS_11_BY_11 = 2;
	protected static final int OSP_ENVIRONMENT_STAR_11_BY_11 = 3;
	protected static final int OSP_ENVIRONMENT_COPWIN_BUILDING_EXAMPLE = 4;
	protected static final int OSP_ENVIRONMENT_PLUS_5_BY_5 = 5;
	protected static final int OSP_ENVIRONMENT_STAR_5_BY_5 = 6;
	
	protected static final int OSP_CURRENT_REGION_POD_30 = 1;
	protected static final int OSP_CURRENT_REGION_POD_60 = 2;
	protected static final int OSP_CURRENT_REGION_POD_90 = 3;
	protected static final int OSP_CURRENT_REGION_POD_99 = 4;
	protected static final int OSP_CURRENT_REGION_POD_80 = 5;
	
	protected static final int OSP_MOTION_MODEL_UNIFORM = 1;
	protected static final int OSP_MOTION_MODEL_STATIONARY = 2;
	protected static final int OSP_MOTION_MODEL_STAY_30 = 3;
	protected static final int OSP_MOTION_MODEL_STAY_60 = 4;
	protected static final int OSP_MOTION_MODEL_STAY_90 = 5;
	
	
	int mId;
	int mT;
	int mN;
	int mQ;
	Graph<Integer,Integer> mEnv;
	int soPosition;
	int suPosition;
	int mMotionModelType;
	double mCurrentRegionPod;
	
	public InstanceConfig() {
		setToInstanceId(0);
	}
	
	InstanceConfig(int id) {
		setToInstanceId(id);
	}
	
	/**
	 * Generate an Optimal Search Path (OSP) problem instance.
	 * 
	 * @param id the Optimal Search Path (OSP) problem identifier
	 */
	public void setToInstanceId(int id) {
		mId = id;
		if(id == 0) {
			mEnv = (new ToyGraphGenerator()).generateCycle(4);
			mT = 4;
			mN = mEnv.getVertexCount();
			mQ = 1;
			soPosition = 0;
			suPosition = 3;
			mMotionModelType = OSP_MOTION_MODEL_UNIFORM;
			mCurrentRegionPod = 0.8;

			return;
		}
		
		// The last 3 digits of the id specify to the total number of time steps
		mT = 0;
		int tmp_id = mId;
		int mod_10;
		for(int i = 0; i < 3; ++i) {
			mod_10 = tmp_id % 10;
			tmp_id = (tmp_id - mod_10) / 10;
			mT += mod_10 * (int) Math.pow(10.0,(double) i);
		}
		
		// The third digit is the motion model type
		mod_10 = tmp_id % 10;
		tmp_id = (tmp_id - mod_10) / 10;
		mMotionModelType = mod_10;
		
		// The second digit is the probability of detection at the searcher's position
		mod_10 = tmp_id % 10;
		tmp_id = (tmp_id - mod_10) / 10;
		switch (mod_10) {
	    	case OSP_CURRENT_REGION_POD_30:  
	    		mCurrentRegionPod = 0.3;
	    		break;
	    	case OSP_CURRENT_REGION_POD_60:  
	    		mCurrentRegionPod = 0.6;
	    		break;
	    	case OSP_CURRENT_REGION_POD_80:  
	    		mCurrentRegionPod = 0.8;
	    		break;
	    	case OSP_CURRENT_REGION_POD_90:  
	    		mCurrentRegionPod = 0.9;
	    		break;
	    	case OSP_CURRENT_REGION_POD_99:  
	    		mCurrentRegionPod = 0.99;
	    		break;
	    	default: 
	    		System.out.println("Warning. The instance id is invalid. The pod digit is invalid.");
	    		break;
	    }
		
		// The first digit is the environment (accessibility graph) type
		mod_10 = tmp_id % 10;
		tmp_id = (tmp_id - mod_10) / 10;
		
		ToyGraphGenerator generator = new ToyGraphGenerator();
		switch (mod_10) {
	    	case OSP_ENVIRONMENT_CAMPUS:  
	    		mEnv = generator.generateLavalCampusMapMini();
				mN = mEnv.getVertexCount();
				soPosition = 16;
				suPosition = 0;
	    		break;
	    	case OSP_ENVIRONMENT_PLUS_11_BY_11:  
	    		int grid_width_plus = 11;
				int grid_height_plus = 11;
				mEnv = generator.generatePlusGrid(grid_width_plus, grid_height_plus);
				mN = mEnv.getVertexCount();
				soPosition = (int) Math.floor((double) (grid_width_plus*grid_height_plus) / 2.0);
				suPosition = 0;
	    		break;
	    	case OSP_ENVIRONMENT_STAR_11_BY_11:  
	    		int grid_width_star = 11;
				int grid_height_star = 11;
				mEnv = generator.generateStarGrid(grid_width_star, grid_height_star);
				mN = mEnv.getVertexCount();
				soPosition = (int) Math.floor((double) (grid_width_star*grid_height_star) / 2.0);
				suPosition = 0;
	    		break;
	    	case OSP_ENVIRONMENT_COPWIN_BUILDING_EXAMPLE:  
	    		mEnv = generator.generateCopwinBuildingExample();
				mN = mEnv.getVertexCount();
				soPosition = 4;
				suPosition = 3;
	    		break;
	    	case OSP_ENVIRONMENT_PLUS_5_BY_5:  
	    		grid_width_plus = 5;
				grid_height_plus = 5;
				mEnv = generator.generatePlusGrid(grid_width_plus, grid_height_plus);
				mN = mEnv.getVertexCount();
				soPosition = (int) Math.floor((double) (grid_width_plus*grid_height_plus) / 2.0);
				suPosition = 0;
	    		break;
	    	case OSP_ENVIRONMENT_STAR_5_BY_5:  
	    		grid_width_star = 5;
				grid_height_star = 5;
				mEnv = generator.generateStarGrid(grid_width_star, grid_height_star);
				mN = mEnv.getVertexCount();
				soPosition = (int) Math.floor((double) (grid_width_star*grid_height_star) / 2.0);
				suPosition = 0;
	    		break;
	    	default: 
	    		System.out.println("Warning. The instance id is invalid. The environment digit is invalid.");
	    		break;
	    }
		
		// Default for all instances
		mQ = 1; // Not used by OSP problem instances
	}
	
	public double[] createPocDistribution() {
		double[] poc_dist = new double[getN()];
		
		assert ((getSoPosition() > 0) && (getSoPosition() < getN()));
			
		poc_dist = new double[getN()];
		
		for(int i = 0; i < getN() ; ++i) {
			if(i == getSoPosition()) {
				poc_dist[i] = 1.0;
			} else {
				poc_dist[i] = 0.0;
			}
		}
		
		return poc_dist;
	}
	
	public double[][] createMotionModel() {
		double[][] motionModel = new double[mN][mN];
		if(getMotionModelType() == InstanceConfig.OSP_MOTION_MODEL_STATIONARY) {
			for(int i = 0; i < getN(); ++i) {
				for(int j = 0; j < getN(); ++j) {
					if(i == j) {
						motionModel[i][j] = 1.0;
					} else {
						motionModel[i][j] = 0.0;
					}
				}
			}
		} else if(getMotionModelType() == InstanceConfig.OSP_MOTION_MODEL_UNIFORM) {
			for(int i = 0; i < getN(); ++i) {
				for(int j = 0; j < getN(); ++j) {
					if(getEnv().isNeighbor(i, j)) {
						motionModel[i][j] = 1.0 / (double) mEnv.getNeighborCount(i);
					}
				}
			}
		} else if((getMotionModelType() == InstanceConfig.OSP_MOTION_MODEL_STAY_30)
					|| (getMotionModelType() == InstanceConfig.OSP_MOTION_MODEL_STAY_60)
					|| (getMotionModelType() == InstanceConfig.OSP_MOTION_MODEL_STAY_90)) {
			double stay_prob = 0.0;
			if(getMotionModelType() == InstanceConfig.OSP_MOTION_MODEL_STAY_30) {
				stay_prob = 0.3;
			} else if(getMotionModelType() == InstanceConfig.OSP_MOTION_MODEL_STAY_60) {
				stay_prob = 0.6;
			} else if(getMotionModelType() == InstanceConfig.OSP_MOTION_MODEL_STAY_90) {
				stay_prob = 0.9;
			}
			
			for(int i = 0; i < getN(); ++i) {
				double sum_row = 0.0;
				for(int j = 0; j < getN(); ++j) {
					if(getEnv().findEdge(i, j) != null) {
						if(i == j) {
							if(mEnv.findEdge(i, i) != null) {
								motionModel[i][j] = stay_prob;
							}
						} else {
							if(mEnv.findEdge(i, i) != null) {
								motionModel[i][j] = (1.0 - stay_prob) / (double) (getEnv().getNeighborCount(i) - 1);
							} else {
								motionModel[i][j] = 1.0 / (double) (getEnv().getNeighborCount(i));
							}
						}
					}
					sum_row += motionModel[i][j];
				}
			}
		}
		return motionModel;
	}
	
	public double getCurrentRegionPod() {
		return mCurrentRegionPod;
	}
	
	public int getId() {
		return mId;
	}
	
	public Graph<Integer,Integer> getEnv() {
		return mEnv;
	}
	
	public int getMotionModelType() {
		return mMotionModelType;
	}
	
	public int getN() {
		return mN;
	}
	
	public int getQ() {
		return mQ;
	}
	
	public int getT() {
		return mT;
	}
	
	public int getSoPosition() {
		return soPosition;
	}
	
	public int getSuPosition() {
		return suPosition;
	}
}

//End of file
