/*
 * Copyright (C) 2010  Michael Morin
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the  v. 3.0  or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the Academic Free License ("AFL")
 * v. 3.0  for more details.
 *
 * You should have received a copy of the Academic Free License ("AFL") v. 3.0
 * along with this program.
 *
 */

/*
 * From the author:
 * This software uses the following third party libraries:
 *  - Choco Solver 2.1.3 (http://www.emn.fr/z-info/choco-solver)
 *  - Jung 2.0.1 (http://jung.sourceforge.net)
 * and is related to the following publication:
 *  M. Morin, A.P. Papillon, F. Laviolette, I. Abi-Zeid, and C.G. Quimper, 
 *  "Constraint Programming for Path Planning with Uncertainty: Solving the 
 *  Optimal Search Path problem," in Proceedings of the 18th Conference on 
 *  Principles and Practice of Constraint Programming, Qubec, Qc, Canada, 
 *  2012, pp. 988-1003.
 * 
 * Enhancements and comments are welcomed. Please let us know!  
 *
 * Best regards,
 *
 * Michael
 * http://www.michaelmorin.info
 */

/** 
 * @file OspSolution.java
 * @brief The Optimal Search Path (OSP) solution instance class.
 */

package ospCp2012;

/**
 * @class OspSolution
 * @version 1.0
 * @author Michael Morin
 * @date Created on: 2011-03-25
 * @date Modified on: 2012-03-22
 *  
 * @brief  class OspSolution
 *
 * This class is the container for an Optimal Search Path (OSP) problem 
 * solution. 
 * <p>
 * For a given problem instance, objects from this class are used to store  
 * the solution path and its Cumulative Overall probability of Success (COS) 
 * value (i.e., its objective value).
 * 
 */
public class OspSolution extends Solution {

	int[] mPath;
	int[][] mAllocation;
	int mT;
	int mN;
	
	OspSolution() {
		mPath = null;
		mAllocation = null;
	}
	
	OspSolution(int regionsN, int timeT) {
		init(regionsN, timeT);
	}
	
	public void init(int regionsN, int timeT) {
		assert (timeT > 0);
		assert (regionsN > 0);
		
		if(mPath != null) {
			mPath = null;
		}
		
		mT = timeT;
		mN = regionsN;
		
		if(mAllocation != null) {
			for(int i = 0; i < mAllocation.length; ++i) {
				mAllocation[i] = null;
			}
			mAllocation = null;
		}
		
		mPath = new int[timeT + 1];
		mAllocation = new int[timeT + 1][regionsN];
		
		for(int i = 0; i < mAllocation.length; ++i) {
			for(int j = 0; j < mAllocation[i].length; ++j) {
				mAllocation[i][j] = 0;
			}
			mPath[i] = 0;
		}
		
		super.init();
	}
	
	public int getSuPosition(int timeStep) {
		assert (timeStep >= 0);
		assert (timeStep <= mT);

		return mPath[timeStep];
	}
	
	public int getSuEffortAllocation(int timeStep, int observedRegion) {
		assert (timeStep >= 0);
		assert (timeStep <= mT);
		assert (observedRegion >= 0);
		assert (observedRegion < mN);
		
		return mAllocation[timeStep][observedRegion];
	}
	
	public void setSuPosition(int timeStep, int suRegion) {
		assert (timeStep >= 0);
		assert (timeStep <= mT);
		assert (suRegion >= 0);
		assert (suRegion < mN);
		
		mPath[timeStep] = suRegion;
	}
	
	public void setSuEffortAllocation(int timeStep, int observedRegion, int effortAmount) {
		assert (timeStep >= 0);
		assert (timeStep <= mT);
		assert (observedRegion >= 0);
		assert (observedRegion < mN);
		assert (effortAmount >= 0);
		
		mAllocation[timeStep][observedRegion] = effortAmount;
	}
	
	/**
	 * Evaluate the Cumulative Overall probability of Success (COS) 
	 * criterion.
	 * 
	 * @param init_poc the initial probability of containment (poc) distribution
	 * @param motion_model the Markovian motion model matrix
	 * @param current_region_pod the probability of detection (pod)
	 */
	public void evaluate(double[] init_poc, double[][] motion_model, double current_region_pod) {

		assert (init_poc.length == mN);
		assert (motion_model.length == mN);
		assert (motion_model[0].length == mN);
		
		double cos_val = 0.0;
		double[] prev_poc_vector = new double[mN];
		double[] curr_poc_vector = new double[mN];
		double[] swap_poc_vector = null;
		for(int r = 0; r < mN; ++r) {
			prev_poc_vector[r] = init_poc[r];
		}
		
		for(int t = 1; t < mT + 1; ++t) {
			for(int r = 0; r < mN; ++r) {
				curr_poc_vector[r] = 0.0;
				for(int s = 0; s < mN; ++s) {
					if(mAllocation[t][s] > 0) {
						curr_poc_vector[r] += motion_model[s][r] * (prev_poc_vector[s] - prev_poc_vector[s] * current_region_pod);
					} else {
						curr_poc_vector[r] += motion_model[s][r] * prev_poc_vector[s];
					}
				}
				if(mAllocation[t][r] > 0) {
					cos_val +=  prev_poc_vector[r] * current_region_pod;
				}
			}

			swap_poc_vector = prev_poc_vector;
			prev_poc_vector = curr_poc_vector;
			curr_poc_vector = swap_poc_vector;
		}		
		
		this.setObjectiveValue(cos_val);
	}
	
	public void resetSearchPlan() {
		for(int t = 0; t < mT + 1; ++t) {
			mPath[t] = 0;
		}

		for(int t = 0; t < mT + 1; ++t) {
			for(int r = 0; r < mN; ++r) {
				mAllocation[t][r] = 0;
			}
		}
	}

}

// End of file
